TEXT_LIST_URL = "https://raw.githubusercontent.com/OpenPecha-dev/editable-text/main/t_text_list.json"
KANGYUR_ARCHIVE_ID = "870326402308432ba173ee3d9043224a"
TENGYUR_ARCHIVE_ID = "187ed94f85154ea5b1ac374a651e1770"



import copy
from pathlib import Path

from openpecha.blupdate import Blupdate, update_ann_layer
from openpecha.utils import download_pecha
from openpecha.utils import dump_yaml, load_yaml

from pedurma.texts import get_pecha_paths, serialize_text_obj
from pedurma.utils import from_yaml


def get_old_vol(pecha_opf_path, pecha_id, text_vol_span):
    """Generate old base text in which text is located

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id 
        text_vol_span (list): list of volume ids in which text is located

    Returns:
        dict: volid as key and base as value
    """
    old_vols = {}
    for vol_id in text_vol_span:
        old_vols[vol_id] = Path(
            f"{pecha_opf_path}/{pecha_id}.opf/base/{vol_id}.txt"
        ).read_text(encoding="utf-8")
    return old_vols


def get_old_text_base(old_pecha_idx, old_vol_base, text_id, text_vol_num):
    """Return old text base

    Args:
        old_pecha_idx (dict): old index
        old_vol_base (dict): old vol base and its vol id
        text_id (str): text id
        text_vol_num (str): text vol id

    Returns:
        str: text basetext in that volume
    """
    text_span = old_pecha_idx["annotations"][text_id]["span"]
    for vol_span in text_span:
        if vol_span["vol"] == text_vol_num:
            return old_vol_base[vol_span["start"] : vol_span["end"]+1]
    return ""


def get_new_vol(old_vols, old_pecha_idx, text_obj):
    """Return new base text by replacing updated text from text object

    Args:
        old_vols (dict): old vol and its id
        old_pecha_idx (dict): old pecha index
        text_obj (obj): text object

    Returns:
        dict: new basetext and its id
    """
    new_vols = {}
    new_text = serialize_text_obj(text_obj)
    for vol_id, new_text_base in new_text.items():
        vol_num = int(vol_id[1:])
        old_vol_base = old_vols[vol_id]
        old_text_base = get_old_text_base(
            old_pecha_idx, old_vol_base, text_obj.id, vol_num
        )
        old_text_base = old_text_base.strip()
        new_text_base = new_text_base.strip()
        new_vol_base = old_vol_base.replace(old_text_base, new_text_base)
        new_vols[vol_id] = new_vol_base
    return new_vols


def get_text_vol_span(pecha_idx, text_uuid):
    """Return list of volume ids in which text span

    Args:
        pecha_idx (dict): pecha index
        text_uuid (uuid): text uuid

    Returns:
        list: vol ids
    """
    text_vol_span = []
    for span in pecha_idx["annotations"][text_uuid]["span"]:
        vol_num = span["vol"]
        text_vol_span.append(f"v{int(vol_num):03}")
    return text_vol_span


def update_base(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update base text using text obj

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for vol_id, new_vol_base in new_vols.items():
        Path(f"{pecha_opf_path}/{pecha_id}.opf/base/{vol_id}.txt").write_text(
            new_vol_base, encoding="utf-8"
        )
        print(f"INFO: {vol_id} base updated..")


def get_old_layers(pecha_opf_path, pecha_id, vol_id):
    """Return all the layers belonging in volume

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        vol_id (str): volume id

    Returns:
        dict: layer name as key and layer annotations as value
    """
    old_layers = {}
    layer_paths = list(
        Path(f"{pecha_opf_path}/{pecha_id}.opf/layers/{vol_id}").iterdir()
    )
    for layer_path in layer_paths:
        layer_name = layer_path.stem
        layer_content = from_yaml(layer_path)
        old_layers[layer_name] = layer_content
    return old_layers


def update_layer(pecha_opf_path, pecha_id, vol_id, old_layers, updater):
    """Update particular layers belonging in given volume id 

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        vol_id (str): volume id
        old_layers (dict): layer name as key and annotations as value
        updater (obj): updater object
    """
    for layer_name, old_layer in old_layers.items():
        if layer_name is not "Pagination":
            update_ann_layer(old_layer, updater)
            new_layer_path = Path(f"{pecha_opf_path}/{pecha_id}.opf/layers/{vol_id}/{layer_name}.yml")
            dump_yaml(old_layer, new_layer_path)
            print(f'INFO: {vol_id} {layer_name} has been updated...')
    
def update_old_layers(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update all the layers related to text object

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for (vol_id, old_vol_base), (_, new_vol_base) in zip(
        old_vols.items(), new_vols.items()
    ):
        updater = Blupdate(old_vol_base, new_vol_base)
        old_layers = get_old_layers(pecha_opf_path, pecha_id, vol_id)
        update_layer(pecha_opf_path, pecha_id, vol_id, old_layers, updater)


def update_other_text_index(old_pecha_idx, text_id, cur_vol_offset, vol_num):
    check_flag = False
    for text_uuid, text in old_pecha_idx["annotations"].items():
        if check_flag:
            for vol_walker, vol_span in enumerate(text["span"]):
                if vol_span["vol"] == vol_num:
                    old_pecha_idx["annotations"][text_uuid]["span"][vol_walker][
                        "start"
                    ] += cur_vol_offset
                    old_pecha_idx["annotations"][text_uuid]["span"][vol_walker][
                        "end"
                    ] += cur_vol_offset
                elif vol_span["vol"] > vol_num:
                    return old_pecha_idx
        if text_uuid == text_id:
            check_flag = True
    return old_pecha_idx


def update_index(pecha_opf_path, pecha_id, text_obj, old_pecha_idx):
    """Update pecha index according to text obj content

    Args:
        pecha_opf_path (str): pecha opf path
        pecha_id (str): pecha id
        text_obj (obj): text object
        old_pecha_idx (dict): old pecha index

    Returns:
        dict: new pecha index
    """
    text_vol_span = get_text_vol_span(old_pecha_idx, text_obj.id)
    old_vols = get_old_vol(pecha_opf_path, pecha_id, text_vol_span)
    new_vols = get_new_vol(old_vols, old_pecha_idx, text_obj)
    for (vol_id, old_vol_base), (_, new_vol_base) in zip(
        old_vols.items(), new_vols.items()
    ):
        check_next_text = True
        vol_num = int(vol_id[1:])
        cur_vol_offset = len(new_vol_base) - len(old_vol_base)
        if cur_vol_offset != 0:
            for vol_walker, vol_span in enumerate(
                old_pecha_idx["annotations"][text_obj.id]["span"]
            ):
                if vol_span["vol"] == vol_num:
                    old_pecha_idx["annotations"][text_obj.id]["span"][vol_walker][
                        "end"
                    ] += cur_vol_offset
                elif vol_span["vol"] > vol_num:
                    check_next_text = False
                    break
            if check_next_text:
                old_pecha_idx = update_other_text_index(
                    old_pecha_idx, text_obj.id, cur_vol_offset, vol_num
                )
    return old_pecha_idx

def update_page_span(page, prev_page_end, old_page_ann):
    new_page_len = len(page.content)
    new_page_end = prev_page_end + new_page_len
    old_page_ann["span"]["start"] = prev_page_end
    old_page_ann["span"]["end"] = new_page_end
    return old_page_ann, new_page_end+2

def update_note_span(pagination_layer, text, prev_page_end):
    for note in text.notes:
        old_page_ann = pagination_layer["annotations"].get(note.id, {})
        if old_page_ann:
            pagination_layer["annotations"][note.id], prev_page_end = update_page_span(note, prev_page_end, old_page_ann)

def get_pagination_layer(pecha_opf_path, pecha_id, vol_num):
    pagination_path = Path(pecha_opf_path) / f"{pecha_id}.opf" / "layers" / f"v{int(vol_num):03}" / "Pagination.yml"
    pagination_layer = load_yaml(pagination_path)
    return pagination_layer, pagination_path

def update_page_layer(text, pecha_id, pecha_opf_path):
    vol_num = text.pages[0].vol
    pagination_layer, pagination_path = get_pagination_layer(pecha_opf_path, pecha_id, vol_num)
    pagination_annotations = pagination_layer.get("annotations", {})
    prev_page_end = 0
    for page in text.pages:
        if vol_num != page.vol:
            update_note_span(pagination_layer, text, prev_page_end)
            prev_page_end = 0
            vol_num = page.vol
            dump_yaml(pagination_layer, pagination_path)
            pagination_layer, pagination_path = get_pagination_layer(pecha_opf_path, pecha_id, vol_num)
            pagination_annotations = pagination_layer.get("annotations", {})
        old_page_ann = pagination_annotations[page.id]
        pagination_layer["annotations"][page.id], prev_page_end = update_page_span(page, prev_page_end, old_page_ann)
    update_note_span(pagination_layer, text, prev_page_end)
    dump_yaml(pagination_layer, pagination_path)


def save_text(pecha_id, text_obj, pecha_opf_path=None, **kwargs):
    """Update pecha opf according to text object content

    Args:
        pecha_id (str): pecha id
        text_obj (text obj): text object
        pecha_opf_path (str, optional): pecha path. Defaults to None.

    Returns:
        path: pecha opf path
    """
    if not pecha_opf_path:
        pecha_opf_path = download_pecha(pecha_id, **kwargs)
    old_pecha_idx = from_yaml(Path(f"{pecha_opf_path}/{pecha_id}.opf/index.yml"))
    prev_pecha_idx = copy.deepcopy(old_pecha_idx)
    new_pecha_idx = update_index(pecha_opf_path, pecha_id, text_obj, old_pecha_idx)
    update_old_layers(pecha_opf_path, pecha_id, text_obj, prev_pecha_idx)
    update_base(pecha_opf_path, pecha_id, text_obj, prev_pecha_idx)
    update_page_layer(text_obj, pecha_id, pecha_opf_path)
    new_pecha_idx_path = Path(f'{pecha_opf_path}/{pecha_id}.opf/index.yml')
    dump_yaml(new_pecha_idx, new_pecha_idx_path)
    return pecha_opf_path


def get_pedurma_text_mapping(pedurma_text_obj):
    """Pedurma text obj are parse and added pecha path

    Args:
        pedurma_text_obj (obj): pedurma text obj

    Returns:
        dict: ocr engine as key and associated text data as value
    """
    pedurma_text_mapping = {}
    pecha_paths = get_pecha_paths(text_id=pedurma_text_obj.text_id)
    for pecha_src, pecha_path in pecha_paths.items():
        if pecha_src == "namsel":
            text_obj = pedurma_text_obj.namsel
        else:
            text_obj = pedurma_text_obj.google
        pedurma_text_mapping[pecha_src] = {
            "pecha_id": Path(pecha_path).stem,
            "text_obj": text_obj,
            "pecha_path": pecha_path,
        }
    return pedurma_text_mapping


def save_pedurma_text(pedurma_text_obj, pedurma_text_mapping=None):
    """Save changes to respective pedurma opfs according to pedurma text object content

    Args:
        pedurma_text_obj (obj): pedurma text object
        pedurma_text_mapping (dict, optional): pedurma text data mapping. Defaults to None.
    """
    if not pedurma_text_mapping:
        pedurma_text_mapping = get_pedurma_text_mapping(pedurma_text_obj)
    for ocr_engine, pedurma_text in pedurma_text_mapping.items():
        save_text(
            pedurma_text["pecha_id"],
            pedurma_text["text_obj"],
            pedurma_text["pecha_path"],
        )



import re
from pathlib import Path

from openpecha.utils import download_pecha

from pedurma import config
from pedurma.pecha import PedurmaNoteEdit
from pedurma.texts import get_durchen, get_hfml_text, get_link, get_vol_meta
from pedurma.utils import from_yaml, get_pecha_id


def get_durchen_pages(vol_text):
    durchen_pages = {}
    pages = re.split(r"(〔[𰵀-󴉱]?\d+〕)", vol_text)
    pg_ann = ""
    for i, page in enumerate(pages[1:]):
        if i % 2 == 0:
            pg_ann = page
        else:
            durchen_pages[pg_ann] = page
    return durchen_pages


def get_page_num(page_ann):
    pg_pat = re.search(r"(\d+)", page_ann)
    if pg_pat:
        pg_num = pg_pat.group(1)
    else:
        pg_num = None
    return pg_num


def rm_annotations(text, annotations):
    clean_text = text
    for ann in annotations:
        clean_text = re.sub(ann, "", clean_text)
    return clean_text


def get_num(line):
    tib_num = re.sub(r"\W", "", line)
    tib_num = re.sub(r"(\d+?)r", "", tib_num)
    table = tib_num.maketrans("༡༢༣༤༥༦༧༨༩༠", "1234567890", "<r>")
    eng_num = int(tib_num.translate(table))
    return eng_num


def get_durchen_pg_num(clean_page):
    pg_num = 0
    try:
        page_ann = re.findall(r"<p\d+-(\d+)\>", clean_page)
        pg_num = page_ann[-1]
    except Exception:
        pass
    return pg_num


def get_page_refs(page_content):
    refs = re.findall(r"<r.+?>", page_content)
    if refs:
        if len(refs) > 2:
            refs[0] = get_num(refs[0])
            refs[-1] = get_num(refs[-1])
            return (refs[0], refs[-1])
        else:
            refs[0] = get_num(refs[0])
            return (refs[0], "0")
    else:
        return ("0", "0")


def process_page(page_ann, page_content, vol_meta):
    durchen_image_num = get_page_num(page_ann)
    pg_link = get_link(durchen_image_num, vol_meta)
    unwanted_annotations = [r"〔[𰵀-󴉱]?\d+〕", r"\[\w+\.\d+\]", r"<d", r"d>"]
    page_content = rm_annotations(page_content, unwanted_annotations)
    durchen_pg_num = get_durchen_pg_num(page_content)
    pg_ref_first, pg_ref_last = get_page_refs(page_content)
    page_obj = PedurmaNoteEdit(
        image_link=pg_link,
        image_no=durchen_image_num,
        page_no=durchen_pg_num,
        ref_start_page_no=pg_ref_first,
        ref_end_page_no=pg_ref_last,
        vol=vol_meta["volume_number"],
    )
    return page_obj


def get_pages_to_edit(durchen_pages, vol_meta):
    pages_to_edit = []
    for page_ann, page_content in durchen_pages.items():
        pages_to_edit.append(process_page(page_ann, page_content, vol_meta))
    return pages_to_edit


def get_pedurma_edit_notes(hfml_text, text_meta):
    pedurma_edit_notes = []
    for vol, text_content in hfml_text.items():
        vol_meta = get_vol_meta(vol, text_meta)
        durchen = get_durchen(text_content)
        durchen_pages = get_durchen_pages(durchen)
        pedurma_edit_notes += get_pages_to_edit(durchen_pages, vol_meta)
    return pedurma_edit_notes


def get_pedurma_text_edit_notes(text_id):
    pecha_id = get_pecha_id(text_id)
    pecha_path = download_pecha(pecha_id, needs_update=False)
    meta_data = from_yaml(Path(f"{pecha_path}/{pecha_id}.opf/meta.yml"))
    hfmls = get_hfml_text(pecha_path, text_id)
    pedurma_edit_notes = get_pedurma_edit_notes(hfmls, meta_data)
    return pedurma_edit_notes



import io
import json
import platform
import re
import stat
import subprocess
import tempfile
import zipfile
from pathlib import Path
from uuid import uuid4

import requests
import yaml

from pedurma import config

PLATFORM_TYPE = platform.system()
BASE_DIR = Path.home() / ".antx"


def get_unique_id():
    return uuid4().hex


def get_pages(vol_text):
    result = []
    pg_text = ""
    pages = re.split(r"(〔[𰵀-󴉱]?\d+〕)", vol_text)
    for i, page in enumerate(pages[1:]):
        if i % 2 == 0:
            pg_text += page
        else:
            pg_text += page
            result.append(pg_text)
            pg_text = ""
    return result


def from_yaml(yml_path):
    return yaml.load(yml_path.read_text(encoding="utf-8"), Loader=yaml.CLoader)


def to_yaml(dict_):
    return yaml.dump(dict_, sort_keys=False, allow_unicode=True, Dumper=yaml.CDumper)


def get_pecha_id(text_id):
    if text_id.startswith("T"):
        return config.KANGYUR_ARCHIVE_ID
    elif text_id.startswith():
        return config.TENGYUR_ARCHIVE_ID
    return ""


def get_bin_metadata():
    """Return platfrom_type and binary_name."""
    if "Windows" in PLATFORM_TYPE:
        return "windows", "dmp.exe"
    elif "Drawin" in PLATFORM_TYPE:
        return "macos", "dmp"
    else:
        return "linux", "dmp"


def get_dmp_bin_url(platform_type):
    response = requests.get(
        "https://api.github.com/repos/Esukhia/node-dmp-cli/releases/latest"
    )
    version = response.json()["tag_name"]
    return (
        f"https://github.com/Esukhia/node-dmp-cli/releases/download/{version}/{platform_type}.zip",
        version,
    )


def get_dmp_exe_path():
    out_dir = BASE_DIR / "bin"
    out_dir.mkdir(exist_ok=True, parents=True)

    platform_type, binary_name = get_bin_metadata()
    binary_path = out_dir / binary_name
    if binary_path.is_file():
        return binary_path

    url, version = get_dmp_bin_url(platform_type)
    print(f"[INFO] Downloading node-dmp-cli-{version} ...")
    r = requests.get(url, stream=True, timeout=50)

    # attempt 50 times to download the zip
    check = zipfile.is_zipfile(io.BytesIO(r.content))
    attempts = 0
    while not check and attempts < 50:
        r = requests.get(url, stream=True, timeout=50)
        check = zipfile.is_zipfile(io.BytesIO(r.content))
        attempts += 1

    if not check:
        raise IOError("the .zip file couldn't be downloaded.")
    else:
        # extract the zip in the current folder
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(path=str(out_dir))

    print("[INFO] Download completed!")

    # make the binary executable
    binary_path.chmod(
        binary_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )
    return str(binary_path)


class optimized_diff_match_patch:
    def __init__(self):
        self.binary_path = get_dmp_exe_path()

    @staticmethod
    def _save_text(text1, text2):
        tmpdir = Path(tempfile.gettempdir())
        text1_path = tmpdir / "text1.txt"
        text2_path = tmpdir / "text2.txt"
        text1_path.write_text(text1, encoding="utf-8")
        text2_path.write_text(text2, encoding="utf-8")
        return str(text1_path), str(text2_path)

    @staticmethod
    def _delete_text(text1_path, text2_path):
        Path(text1_path).unlink()
        Path(text2_path).unlink()

    @staticmethod
    def _unescape_lr(diffs):
        """Unescape the line-return."""
        for diff_type, diff_text in diffs:
            if "Windows" in PLATFORM_TYPE:
                yield (diff_type, diff_text.replace("\r\\n", "\n"))
            else:
                yield (diff_type, diff_text.replace("\\n", "\n"))

    def diff_main(self, text1, text2):
        text1_path, text2_path = self._save_text(text1, text2)
        process = subprocess.Popen(
            [str(self.binary_path), "diff", text1_path, text2_path],
            stdout=subprocess.PIPE,
        )
        stdout = process.communicate()[0]
        diffs = json.loads(stdout)
        diffs = self._unescape_lr(diffs)
        self._delete_text(text1_path, text2_path)
        return diffs



from typing import List, Optional

from pydantic import AnyHttpUrl, BaseModel


# Shared properties
class PageBase(BaseModel):
    id: str
    page_no: int
    content: str
    name: str
    vol: str
    image_link: Optional[str] = None


class Page(PageBase):
    note_ref: Optional[str]


class NotesPage(PageBase):
    pass


class PageWithNote(BaseModel):
    text_id: str
    page_id: str
    content: str
    page_image_link: AnyHttpUrl
    note: str
    note_image_link: AnyHttpUrl


class Text(BaseModel):
    id: str
    pages: List[Page]
    notes: Optional[List[NotesPage]]


class PedurmaText(BaseModel):
    text_id: str
    namsel: Text
    google: Text


class PedurmaPreviewPage(BaseModel):
    content: str


class PedurmaNoteEdit(BaseModel):
    image_link: str
    image_no: int
    page_no: int
    ref_start_page_no: str
    ref_end_page_no: str
    vol: int


class EditorContent(BaseModel):
    content: str


class BaseLayer(BaseModel):
    content: str


class PechaBase(BaseModel):
    id: Optional[str] = None
    title: Optional[str] = None
    img: Optional[AnyHttpUrl] = None


class PechaCreate(PechaBase):
    id: str
    title: str


class PechaUpdate(PechaBase):
    id: str
    title: str


class PechaInDBBase(PechaBase):
    id: str
    title: str
    owner_id: int
    img: Optional[AnyHttpUrl] = None

    class Config:
        orm_mode = True


class Pecha(PechaInDBBase):
    pass


class PechaInDB(PechaInDBBase):
    pass
